from typing import Any

# Credits to https://code.activestate.com/recipes/215912-union-find-data-structure/


class UnionFind:
    def __init__(self) -> None:
        self.weights: dict[int, int] = {}
        self.pointers: dict[int, int] = {}
        self.indexes_to_objects: dict[int, Any] = {}
        self.objects_to_indexes: dict[Any, int] = {}

    def insert_objects(self, objects: Any) -> None:
        for object in objects:
            self.find(object)

    def build_if_not_exists(self, object: Any) -> Any:
        new_index = len(self.objects_to_indexes)

        self.weights[new_index] = 1
        self.objects_to_indexes[object] = new_index
        self.indexes_to_objects[new_index] = object
        self.pointers[new_index] = new_index

        return object

    def find(self, object: Any) -> Any:
        if object not in self.objects_to_indexes:
            return self.build_if_not_exists(object)

        object_indexes = [self.objects_to_indexes[object]]
        object_pointers = self.pointers[object_indexes[-1]]

        while object_pointers != object_indexes[-1]:
            object_indexes.append(object_pointers)
            object_pointers = self.pointers[object_pointers]

        for index in object_indexes:
            self.pointers[index] = object_pointers

        return self.indexes_to_objects[object_pointers]

    def union(self, object_a: Any, object_b: Any) -> None:
        object_a_set = self.find(object_a)
        object_b_set = self.find(object_b)

        if object_a_set != object_b_set:
            object_a_indexes = self.objects_to_indexes[object_a_set]
            object_b_indexes = self.objects_to_indexes[object_b_set]
            object_a_weights = self.weights[object_a_indexes]
            object_b_weights = self.weights[object_b_indexes]

            if object_a_weights < object_b_weights:
                (
                    object_a_set,
                    object_b_set,
                    object_a_indexes,
                    object_b_indexes,
                    object_a_weights,
                    object_b_weights,
                ) = (
                    object_b_set,
                    object_a_set,
                    object_b_indexes,
                    object_a_indexes,
                    object_b_weights,
                    object_a_weights,
                )

            self.weights[object_a_indexes] = object_a_weights + object_b_weights
            del self.weights[object_b_indexes]
            self.pointers[object_b_indexes] = object_a_indexes



import numpy as np
import pytest

from topologyx.clustering import ClusterGenerator, ClusterStructure


@pytest.fixture
def simple_2d_data() -> np.ndarray:
    """Simple 2D dataset for basic tests."""
    return np.array([[0, 0], [1, 0], [0, 1], [1, 1], [5, 5], [6, 5], [5, 6], [6, 6]])


@pytest.fixture
def blobs_data() -> tuple[np.ndarray, np.ndarray]:
    """Generate blob dataset with known structure."""
    generator = ClusterGenerator(
        structure=ClusterStructure.BLOBS, n_samples=300, randomize=42
    )
    x, y = generator.generate()
    return x, y


@pytest.fixture
def moons_data() -> tuple[np.ndarray, np.ndarray]:
    """Generate moons dataset with known structure."""
    generator = ClusterGenerator(
        structure=ClusterStructure.MOONS, n_samples=300, randomize=42
    )
    x, y = generator.generate()
    return x, y


@pytest.fixture
def circles_data() -> tuple[np.ndarray, np.ndarray]:
    """Generate circles dataset with known structure."""
    generator = ClusterGenerator(
        structure=ClusterStructure.CIRCLES, n_samples=300, randomize=42
    )
    x, y = generator.generate()
    return x, y



from typing import Any

import gudhi
import matplotlib.pyplot as plt
import numpy as np


def functionize(vector: np.ndarray, descriptor: tuple[float, float]) -> np.ndarray:
    return np.vectorize(
        lambda x: 1 if (x > descriptor[0]) and (x < descriptor[1]) else 0
    )(vector)


def build_betti_curve(
    persistence: Any,
    minimum: float | None = None,
    maximum: float | None = None,
    n_points: int = 100,
) -> np.ndarray:
    if minimum is None:
        minimum = np.min(persistence)

    if maximum is None:
        maximum = np.max(persistence)

    linspace = np.linspace(minimum, maximum, num=n_points)  # type: ignore

    return np.zeros(n_points) + sum(
        functionize(linspace, component) for component in persistence
    )


def build_persistence_landscape(
    persistence: Any,
    minimum: float | None = None,
    maximum: float | None = None,
    n_landscapes: int = 5,
    n_points: int = 100,
) -> np.ndarray:
    if minimum is None:
        minimum = np.min(persistence)

    if maximum is None:
        maximum = np.max(persistence)

    support = np.zeros((n_landscapes, n_points))
    linspace = np.linspace(minimum, maximum, num=n_points)  # type: ignore

    for i, value in enumerate(linspace):
        _vector = []

        for pair in persistence:
            birth_radius, death_radius = pair[0], pair[1]

            if (death_radius + birth_radius) / 2.0 <= value <= death_radius:
                _vector.append(death_radius - value)
            elif birth_radius <= value <= (death_radius + birth_radius) / 2.0:
                _vector.append(value - birth_radius)

        _vector.sort(reverse=True)
        vector = np.asarray(_vector)

        for j in range(n_landscapes):
            if j < len(vector):
                support[j, i] = vector[j]

    return support


def _compute_point_weight(point: tuple[float, float], extrema: float) -> float:
    return 0 if point[1] <= 0 else point[1] / extrema


def _compute_gaussian_kernel(
    point: tuple[float, float],
    x_extremas: tuple[float, float],
    y_extremas: tuple[float, float],
    image_size: tuple[int, int],
    variance: float,
) -> np.ndarray:
    image = np.zeros(image_size)

    x_linspace = np.linspace(x_extremas[0], x_extremas[1], image_size[0])
    y_linspace = np.linspace(y_extremas[0], y_extremas[1], image_size[1])

    for i, x in enumerate(x_linspace):
        for j, y in enumerate(y_linspace):
            image[len(y_linspace) - j - 1, i] = (
                _compute_point_weight(point, y_extremas[1])
                * (1.0 / (2 * np.pi * variance))
                * np.exp(
                    -(np.square(x - point[0]) + np.square(y - point[1]))
                    / (2 * variance)
                )
            )

    return image


def build_persistence_image(
    persistence: Any,
    x_extremas: tuple[float, float] | None,
    y_extremas: tuple[float, float] | None,
    image_size: tuple[int, int],
    variance: float,
) -> np.ndarray:
    image = np.zeros(image_size)

    normalized_persistence = persistence.copy()
    normalized_persistence[:, 1] = (
        normalized_persistence[:, 1] - np.sum(normalized_persistence, axis=1) / 2
    )

    if x_extremas is None:
        x_extremas = (
            np.min(normalized_persistence[:, 0]),
            np.max(normalized_persistence[:, 0]),
        )

    if y_extremas is None:
        y_extremas = (
            np.min(normalized_persistence[:, 1]),
            np.max(normalized_persistence[:, 1]),
        )

    for point in normalized_persistence:
        image += _compute_gaussian_kernel(
            point, x_extremas, y_extremas, image_size, variance
        )

    return image


def plot_persistences(upper_persistence: Any, lower_persistence: Any) -> None:
    _, axes = plt.subplots(nrows=2, ncols=2, figsize=(18, 8))

    gudhi.plot_persistence_diagram(upper_persistence, axes=axes[0, 0])  # type: ignore
    gudhi.plot_persistence_barcode(upper_persistence, axes=axes[1, 0])  # type: ignore
    gudhi.plot_persistence_diagram(lower_persistence, axes=axes[0, 1])  # type: ignore
    gudhi.plot_persistence_barcode(lower_persistence, axes=axes[1, 1])  # type: ignore

    plt.tight_layout()
    plt.show()


def plot_betti_curves(upper_levels: np.ndarray, lower_levels: np.ndarray) -> None:
    # sourcery skip: extract-duplicate-method
    plt.figure(figsize=(18, 3))
    plt.subplot(1, 2, 1)
    plt.title("Upper Levels Betti Curve")
    plt.plot(upper_levels)
    plt.xticks([])
    plt.yticks([])
    plt.subplot(1, 2, 2)
    plt.title("Lower Levels Betti Curve")
    plt.plot(lower_levels)
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    plt.show()


def plot_persistence_landscapes(
    upper_levels: np.ndarray,
    lower_levels: np.ndarray,
) -> None:
    # sourcery skip: extract-duplicate-method
    plt.figure(figsize=(18, 3))
    plt.subplot(1, 2, 1)
    plt.title("Upper Levels Persistence Landscapes")
    for component in upper_levels:
        plt.plot(component)
    plt.xticks([])
    plt.yticks([])
    plt.subplot(1, 2, 2)
    plt.title("Lower Levels Persistence Landscapes")
    for component in lower_levels:
        plt.plot(component)
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    plt.show()



import marimo

__generated_with = "0.18.4"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo

    return (mo,)


@app.cell
def _(mo):
    mo.md(r"""
    # ToMaTo Clustering Algorithm

    Implementation of the ToMaTo clustering algorithm, presented by Steve Oudot
    in "Persistence Theory: From Quiver Representations to Data Analysis".

    ## Concept

    The algorithm exploits topological data analysis for clustering. Given a
    two-dimensional dataset, the sublevels filtration of the simplex tree built
    over the density estimate gives, by reconstruction of the connected components,
    the hierarchical relation between the data points.

    By sorting it, we can retrace the density extrema and thus the centroids.
    Visualized persistence gives a good indication of the number of centroids.
    Once determined, the construction of distinct sets through UnionFind becomes possible.

    ## Algorithm

    ![ToMaTo Algorithm](./figures/tomato.png)
    """)
    return


@app.cell
def _():
    import warnings

    warnings.simplefilter("ignore")
    return


@app.cell
def _():
    from topologyx.clustering import (
        ClusterGenerator,
        ClusterStructure,
        TomatoClustering,
    )

    return ClusterGenerator, ClusterStructure, TomatoClustering


@app.cell
def _(mo):
    mo.md(r"""
    ## Example 1: Anisotropic Distribution
    """)
    return


@app.cell
def _(ClusterGenerator, ClusterStructure, TomatoClustering):
    generator_aniso = ClusterGenerator(
        structure=ClusterStructure.ANISOTROPY, randomize=20
    )
    tomato_aniso = TomatoClustering(*generator_aniso.generate())
    _ = tomato_aniso.estimate_clusters(visualize=True)
    _ = tomato_aniso.fit_predict(n_clusters=3, visualize=True)
    return


@app.cell
def _(mo):
    mo.md(r"""
    ## Example 2: Two Moons
    """)
    return


@app.cell
def _(ClusterGenerator, ClusterStructure, TomatoClustering):
    generator_moons = ClusterGenerator(structure=ClusterStructure.MOONS, randomize=18)
    tomato_moons = TomatoClustering(*generator_moons.generate())
    _ = tomato_moons.estimate_clusters(visualize=True)
    _ = tomato_moons.fit_predict(n_clusters=2, visualize=True)
    return


@app.cell
def _(mo):
    mo.md(r"""
    ## Example 3: Blobs
    """)
    return


@app.cell
def _(ClusterGenerator, ClusterStructure, TomatoClustering):
    generator_blobs = ClusterGenerator(structure=ClusterStructure.BLOBS, randomize=28)
    tomato_blobs = TomatoClustering(*generator_blobs.generate())
    _ = tomato_blobs.estimate_clusters(visualize=True)
    _ = tomato_blobs.fit_predict(n_clusters=2, visualize=True)
    return


@app.cell
def _(mo):
    mo.md(r"""
    ## Example 4: Random Distribution
    """)
    return


@app.cell
def _(ClusterGenerator, ClusterStructure, TomatoClustering):
    generator_random = ClusterGenerator(structure=ClusterStructure.RANDOM, randomize=30)
    tomato_random = TomatoClustering(*generator_random.generate())
    _ = tomato_random.estimate_clusters(n_neighbors=10, visualize=True)
    _ = tomato_random.fit_predict(n_clusters=6, tau=0.001, visualize=True)
    return


@app.cell
def _(mo):
    mo.md(r"""
    ## Example 5: Concentric Circles
    """)
    return


@app.cell
def _(ClusterGenerator, ClusterStructure, TomatoClustering):
    generator_circles = ClusterGenerator(
        structure=ClusterStructure.CIRCLES, randomize=60
    )
    tomato_circles = TomatoClustering(*generator_circles.generate())
    _ = tomato_circles.estimate_clusters(n_neighbors=25, visualize=True)
    _ = tomato_circles.fit_predict(n_clusters=2, tau=1.0, visualize=True)
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



import gudhi
import matplotlib.gridspec as gds
import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import gaussian_kde as kde


def plot_persistence(persistence: list[tuple[int, list[tuple[float, float]]]]) -> None:
    reduced_persistence = [
        (dimension, radiuses) for (dimension, radiuses) in persistence if dimension == 0
    ]

    _, axes = plt.subplots(nrows=1, ncols=2, figsize=(18, 4))

    gudhi.plot_persistence_diagram(reduced_persistence, axes=axes[0])  # type: ignore

    gudhi.plot_persistence_barcode(reduced_persistence, axes=axes[1])  # type: ignore

    plt.tight_layout()
    plt.show()


def plot_density(tensor: np.ndarray, n_bins: int, vector_density: np.ndarray) -> None:
    x, y = tensor.T

    u, v = np.mgrid[
        int(x.min()) : int(x.max()) : n_bins * 1j,  # type: ignore
        int(y.min()) : int(y.max()) : n_bins * 1j,  # type: ignore
    ]

    kernel_density = kde(tensor.T)(np.vstack([u.flatten(), v.flatten()]))

    plt.figure(figsize=(18, 10))
    fig = gds.GridSpec(3, 6)

    plt.subplot(fig[0, 0:2])
    plt.title("Data Scatter Plot")
    plt.plot(x, y, "ko")
    plt.xticks([])
    plt.yticks([])

    plt.subplot(fig[0, 2:4])
    plt.title("Gaussian KDE")
    plt.pcolormesh(u, v, kernel_density.reshape(u.shape))
    plt.xticks([])
    plt.yticks([])

    plt.subplot(fig[0, 4:6])
    plt.title("Density Contours")
    plt.pcolormesh(u, v, kernel_density.reshape(u.shape))
    plt.contour(u, v, kernel_density.reshape(u.shape))
    plt.xticks([])
    plt.yticks([])

    axe_0 = plt.subplot(fig[1:3, 0:3], projection="3d")
    axe_0.set_title("Mapped Density over 2D Space")
    axe_0.set_xticks([])
    axe_0.set_yticks([])
    axe_0.set_zticks([])  # type: ignore
    axe_0.scatter(u, v, kernel_density.reshape(u.shape), s=2, c="lightblue")  # type: ignore
    axe_0.set_xlabel("x Coordinate")
    axe_0.set_ylabel("y Coordinate")
    axe_0.set_zlabel("Density Value")  # type: ignore

    axe_1 = plt.subplot(fig[1:3, 3:6], projection="3d")
    axe_1.set_title("Density Estimate over 2D Space")
    axe_1.set_xticks([])
    axe_1.set_yticks([])
    axe_1.set_zticks([])  # type: ignore
    axe_1.scatter(x, y, vector_density.reshape(x.shape), s=2, c="lightgrey")  # type: ignore
    axe_1.set_xlabel("x Coordinate")
    axe_1.set_ylabel("y Coordinate")
    axe_1.set_zlabel("Density Value")  # type: ignore

    plt.tight_layout()
    plt.show()


def plot_clusters(
    values: np.ndarray, clusters: list[list[int]], centroids: list[int]
) -> None:
    plt.figure(figsize=(18, 4))

    plt.subplot(1, 2, 1)
    plt.title("Initial Data")
    plt.scatter(values[:, 0], values[:, 1], c="lightgrey")
    plt.xticks([])
    plt.yticks([])

    plt.subplot(1, 2, 2)
    plt.title("Clustered Data")

    for index, label in enumerate(clusters):
        plt.scatter(values[label, 0], values[label, 1], label=f"Cluster {index}")

    plt.scatter(
        values[centroids, 0],
        values[centroids, 1],
        c="black",
        marker="x",
        label="Centroids",
    )

    plt.legend(loc="best")
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()
    plt.show()



from enum import Enum
from typing import Protocol

import gudhi
import numpy as np
from sklearn.neighbors import KDTree


class SimplexTreeProtocol(Protocol):
    """Protocol defining the interface for gudhi SimplexTree objects."""

    def persistence(self) -> list[tuple[int, tuple[float, float]]]: ...
    def persistence_intervals_in_dimension(
        self, dimension: int
    ) -> list[tuple[float, float]]: ...
    def insert(self, simplex: list[int], filtration: float = 0.0) -> bool: ...
    def set_dimension(self, dimension: int) -> None: ...
    def initialize_filtration(self) -> None: ...


from topologyx.filtrations.utils import (
    build_betti_curve,
    build_persistence_image,
    build_persistence_landscape,
    plot_betti_curves,
    plot_persistence_landscapes,
    plot_persistences,
)


class FiltrationType(Enum):
    DTM = "dtm"
    SIMPLE = "simple"
    SUBLEVEL = "sublevel"


class Filtration:
    def __init__(
        self,
        vector: np.ndarray,
        use_alpha: bool = True,
        leaf_size: int = 30,
    ) -> None:
        self.vector = vector
        self.kd_tree = KDTree(self.vector, leaf_size=leaf_size, metric="euclidean")

        if use_alpha:
            self.simplex = gudhi.AlphaComplex(points=self.vector)  # type: ignore
            self.simplex = self.simplex.create_simplex_tree(max_alpha_square=250.0)

        else:
            self.simplex = gudhi.SimplexTree()  # type: ignore

            if vector.shape[1] == 1:
                for i in np.arange(len(vector)):
                    self.simplex.insert([i], filtration=vector[i])

                for i in np.arange(len(vector) - 1):
                    self.simplex.insert([i, i + 1], filtration=vector[i])

            elif vector.shape[1] == 2:
                for index in range(len(vector)):
                    self.simplex.insert([index], filtration=-vector[index, 1])

                    for neighbor in self.kd_tree.query(
                        [vector[index, :]], 5, return_distance=False
                    )[0][1:]:
                        self.simplex.insert(
                            [index, neighbor],
                            filtration=np.mean(
                                [-vector[index, 1], -vector[neighbor, 1]]
                            ),
                        )

            elif vector.shape[1] == 3:
                for index in range(len(vector)):
                    self.simplex.insert([index], filtration=-vector[index, 2])

                    for neighbor in self.kd_tree.query(
                        [vector[index, :]], 5, return_distance=False
                    )[0][1:]:
                        self.simplex.insert(
                            [index, neighbor],
                            filtration=np.mean(
                                [-vector[index, 2], -vector[neighbor, 2]]
                            ),
                        )

            else:
                raise ValueError("Dimension not supported")

            self.simplex.initialize_filtration()

    def get_density_estimate(self, points: np.ndarray, n_neighbors: int) -> np.ndarray:
        neighbors = self.kd_tree.query(points, n_neighbors, return_distance=True)

        return np.sqrt(np.sum(np.square(neighbors[0]), axis=1) / n_neighbors)

    def get_vertexes(self, n_neighbors: int) -> np.ndarray:
        return self.get_density_estimate(self.vector, n_neighbors)

    def apply_sublevel_filtration(self, n_neighbors: int = 5) -> SimplexTreeProtocol:
        simplex = gudhi.SimplexTree()  # type: ignore
        vertexes = self.get_vertexes(n_neighbors)

        for pair in self.simplex.get_filtration():  # type: ignore
            simplex.insert(
                pair[0],
                filtration=max(pair[1], max(vertexes[i] for i in pair[0])),  # type: ignore
            )

        simplex.set_dimension(self.vector.shape[1])
        simplex.initialize_filtration()
        simplex.persistence()

        return simplex

    def compute_segment_maximum_density(
        self,
        p: np.ndarray,
        q: np.ndarray,
        n_divisions: int,
        n_neighbors: int,
    ) -> float:
        step = (q - p) / float(n_divisions)
        points = np.zeros((n_divisions + 1, len(p)))

        for i in range(n_divisions):
            points[i, :] = p + i * step

        points[n_divisions, :] = q

        return max(self.get_density_estimate(points, n_neighbors))

    def compute_triangle_maximum_density(
        self,
        p: np.ndarray,
        q: np.ndarray,
        r: np.ndarray,
        n_divisions: int,
        n_neighbors: int,
    ) -> float:
        points = []
        for alpha in range(n_divisions):
            for beta in range(n_divisions - alpha):
                gamma = n_divisions - alpha - beta

                points.append(
                    [(alpha * p + beta * q + gamma * r) / float(n_divisions), p, q, r]
                )

        return max(self.get_density_estimate(np.asarray(points), n_neighbors))

    def apply_dtm_filtration(
        self, n_neighbors: int = 5, n_divisions: int = 5
    ) -> SimplexTreeProtocol:
        simplex = gudhi.SimplexTree()  # type: ignore
        vertexes = self.get_vertexes(n_neighbors)

        for pair in self.simplex.get_filtration():  # type: ignore
            if len(pair[0]) == 1:
                simplex.insert(pair[0], filtration=vertexes[pair[0][0]])

            elif len(pair[0]) == 2:
                simplex.insert(
                    pair[0],
                    filtration=self.compute_segment_maximum_density(
                        self.vector[pair[0][0], :],
                        self.vector[pair[0][1], :],
                        n_divisions,
                        n_neighbors,
                    ),
                )

            elif len(pair[0]) == 3:
                simplex.insert(
                    pair[0],
                    filtration=self.compute_triangle_maximum_density(
                        self.vector[pair[0][0], :],
                        self.vector[pair[0][1], :],
                        self.vector[pair[0][2], :],
                        n_divisions,
                        n_neighbors,
                    ),
                )

            else:
                raise ValueError("Dimension not supported")

        simplex.set_dimension(self.vector.shape[1])
        simplex.initialize_filtration()
        simplex.persistence()

        return simplex

    def build_persistence_diagram(
        self, filtration_type: FiltrationType | None = None, dimension: int = 0
    ) -> None:
        simplex_tree: SimplexTreeProtocol = self.simplex  # type: ignore[assignment]

        match filtration_type:
            case FiltrationType.DTM:
                simplex_tree = self.apply_dtm_filtration()
            case FiltrationType.SUBLEVEL:
                simplex_tree = self.apply_sublevel_filtration()
            case _:
                simplex_tree.persistence()

        self.simplex = np.asarray(
            [
                [interval[0], interval[1]]
                for interval in simplex_tree.persistence_intervals_in_dimension(
                    dimension
                )
                if interval[1] < np.inf
            ]
        )

    def betti_curve(
        self,
        minimum: float | None = None,
        maximum: float | None = None,
        n_points: int = 100,
    ) -> np.ndarray:
        return build_betti_curve(
            self.simplex,
            minimum=minimum,
            maximum=maximum,
            n_points=n_points,
        )

    def persistence_landscapes(
        self,
        minimum: float | None = None,
        maximum: float | None = None,
        n_landscapes: int = 10,
        n_points: int = 100,
    ) -> np.ndarray:
        return build_persistence_landscape(
            self.simplex,
            minimum=minimum,
            maximum=maximum,
            n_landscapes=n_landscapes,
            n_points=n_points,
        )

    def persistence_image(
        self,
        x_extremas: tuple[float, float] | None = None,
        y_extremas: tuple[float, float] | None = None,
        image_size: tuple[int, int] = (32, 32),
        variance: float = 1e-8,
    ) -> np.ndarray:
        return build_persistence_image(
            self.simplex,
            x_extremas,
            y_extremas,
            image_size,
            variance,
        )


class Levels:
    def __init__(self, serie: np.ndarray) -> None:
        self.upper_simplex = gudhi.SimplexTree()  # type: ignore
        self.lower_simplex = gudhi.SimplexTree()  # type: ignore

        for index in np.arange(len(serie)):
            self.upper_simplex.insert([index], filtration=serie[index])
            self.lower_simplex.insert([index], filtration=-serie[index])

        for index in np.arange(len(serie) - 1):
            self.upper_simplex.insert([index, index + 1], filtration=serie[index])
            self.lower_simplex.insert([index, index + 1], filtration=-serie[index])

        self.upper_simplex.initialize_filtration()
        self.lower_simplex.initialize_filtration()

    def build_persistence_diagram(
        self, visualize: bool = False
    ) -> tuple[
        np.ndarray,
        np.ndarray,
    ]:
        upper_persistence = self.upper_simplex.persistence()
        lower_persistence = self.lower_simplex.persistence()

        if visualize:
            plot_persistences(upper_persistence, lower_persistence)

        upper_persistence = np.asarray(
            [
                [component[1][0], component[1][1]]
                for component in upper_persistence
                if component[1][1] < np.inf
            ]
        )

        lower_persistence = np.asarray(
            [
                [component[1][0], component[1][1]]
                for component in lower_persistence
                if component[1][1] < np.inf
            ]
        )

        return upper_persistence, lower_persistence

    def betti_curves(
        self,
        upper_minimum: float | None = None,
        upper_maximum: float | None = None,
        lower_minimum: float | None = None,
        lower_maximum: float | None = None,
        n_points: int = 100,
        visualize: bool = False,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
    ]:
        upper_persistence, lower_persistence = self.build_persistence_diagram(
            visualize=visualize
        )

        upper_curve = build_betti_curve(
            upper_persistence,
            minimum=upper_minimum,
            maximum=upper_maximum,
            n_points=n_points,
        )
        lower_curve = build_betti_curve(
            lower_persistence,
            minimum=lower_minimum,
            maximum=lower_maximum,
            n_points=n_points,
        )

        if visualize:
            plot_betti_curves(upper_curve, lower_curve)

        return upper_curve, lower_curve

    def persistence_landscapes(
        self,
        upper_minimum: float | None = None,
        upper_maximum: float | None = None,
        lower_minimum: float | None = None,
        lower_maximum: float | None = None,
        n_landscapes: int = 5,
        n_points: int = 100,
        visualize: bool = False,
    ) -> tuple[
        np.ndarray,
        np.ndarray,
    ]:
        upper_persistence, lower_persistence = self.build_persistence_diagram(
            visualize=visualize
        )

        upper_landscapes = build_persistence_landscape(
            upper_persistence,
            minimum=upper_minimum,
            maximum=upper_maximum,
            n_landscapes=n_landscapes,
            n_points=n_points,
        )
        lower_landscapes = build_persistence_landscape(
            lower_persistence,
            minimum=lower_minimum,
            maximum=lower_maximum,
            n_landscapes=n_landscapes,
            n_points=n_points,
        )

        if visualize:
            plot_persistence_landscapes(upper_landscapes, lower_landscapes)

        return upper_landscapes, lower_landscapes



from typing import Any

from keras import initializers, layers, ops


class SilhouetteLayer(layers.Layer):  # type: ignore
    def __init__(self, dimension: int, **kwargs: Any) -> None:
        self.dimension = dimension

        super(SilhouetteLayer, self).__init__(**kwargs)

    def build(self, input_shape: tuple[int, ...]) -> None:
        self.kernel = self.add_weight(
            name="kernel",
            shape=(1, input_shape[-2]),
            initializer=initializers.Constant(value=1.0),
            trainable=True,
        )

        super(SilhouetteLayer, self).build(input_shape)

    def call(self, inputs: Any) -> Any:
        return ops.reshape(
            ops.sum(ops.dot(self.kernel, inputs), axis=0),
            (-1, inputs.get_shape()[-1], 1),
        )

    def compute_output_shape(self, _: Any) -> tuple[None, int, int]:
        return (None, self.dimension, 1)



from typing import Any

from keras import layers

from topologyx.channels.silhouette import SilhouetteLayer


def betti_channel(input: Any, args: dict[str, Any], dropout: float = 0.3) -> Any:
    channel = layers.Reshape((input._keras_shape[1], 1))(input)
    channel = layers.Conv1D(64, 10, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Dropout(dropout)(channel)
    channel = layers.Conv1D(128, 6, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Conv1D(128, 6, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Conv1D(128, 4, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Dropout(dropout)(channel)
    channel = layers.GlobalAveragePooling1D()(channel)

    return channel


def silhouette_channel(input: Any, args: dict[str, Any], dropout: float = 0.3) -> Any:
    channel = SilhouetteLayer(int(input._keras_shape[-1]))(input)

    channel = layers.Conv1D(64, 10, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Dropout(dropout)(channel)
    channel = layers.Conv1D(128, 6, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Conv1D(128, 6, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Conv1D(128, 4, **args)(channel)
    channel = layers.PReLU()(channel)
    channel = layers.Dropout(dropout)(channel)
    channel = layers.GlobalAveragePooling1D()(channel)

    return channel



import marimo

__generated_with = "0.18.4"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo

    return (mo,)


@app.cell
def _(mo):
    mo.md(
        r"""
        # TDA on 3D Shapes

        This notebook demonstrates the use of Topological Data Analysis (TDA)
        on three-dimensional point clouds. We use height as the filtration function
        to analyze the topological features of a 3D hand shape.
        """
    )
    return


@app.cell
def _():
    import warnings

    warnings.simplefilter("ignore")
    return (warnings,)


@app.cell
def _(mo):
    mo.md(r"""## Load the 3D Point Cloud""")
    return


@app.cell
def _():
    import numpy as np

    def load_shape_file(filepath: str) -> np.ndarray:
        cloud = []
        with open(filepath, "r") as stream:
            while len(coordinates := stream.readline().split(" ")) == 3:
                cloud.append(np.asarray(coordinates).astype("float"))

        return np.vstack(cloud)

    cloud = load_shape_file("./samples/hand.off")
    return cloud, load_shape_file, np


@app.cell
def _(mo):
    mo.md(
        r"""
        ## Build Persistence Diagram

        Extract the connected components (dimension 0) from the persistence diagram.
        """
    )
    return


@app.cell
def _():
    from topologyx.filtrations import Filtration, FiltrationType

    return Filtration, FiltrationType


@app.cell
def _(Filtration, FiltrationType, cloud):
    filtration = Filtration(cloud, use_alpha=False)
    filtration.build_persistence_diagram(
        filtration_type=FiltrationType.SIMPLE, dimension=0
    )
    return (filtration,)


@app.cell
def _(mo):
    mo.md(r"""## Compute Betti Curve""")
    return


@app.cell
def _(filtration):
    betti_curves = filtration.betti_curve()
    return (betti_curves,)


@app.cell
def _(mo):
    mo.md(r"""## Compute Persistence Landscapes""")
    return


@app.cell
def _(filtration):
    persistence_landscapes = filtration.persistence_landscapes()
    return (persistence_landscapes,)


@app.cell
def _(mo):
    mo.md(r"""## Visualization""")
    return


@app.cell
def _(betti_curves, cloud, filtration, persistence_landscapes):
    import gudhi
    import matplotlib.pyplot as plt

    diagram = [(0, tuple(e)) for e in filtration.simplex]

    fig, axes = plt.subplots(nrows=4, ncols=2, figsize=(18, 14))
    grid = axes[0, 0].get_gridspec()

    for axe in axes[0:, 0]:
        axe.remove()

    axe_0 = fig.add_subplot(grid[0:, 0], projection="3d")
    axe_0.set_box_aspect(aspect=(1.5, 1.5, 2.75))  # type: ignore
    axe_0.set_title("Mapped 3 Dimensional Data Cloud")
    axe_0.set_xticks([])
    axe_0.set_yticks([])
    axe_0.set_zticks([])  # type: ignore
    axe_0.scatter(cloud[:, 0], cloud[:, 1], cloud[:, 2], s=2, c="coral")  # type: ignore
    axe_0.set_xlabel("x Coordinate")
    axe_0.set_ylabel("y Coordinate")
    axe_0.set_zlabel("z Coordinate")  # type: ignore

    gudhi.plot_persistence_diagram(diagram, axes=axes[0, 1])  # type: ignore

    gudhi.plot_persistence_barcode(diagram, axes=axes[1, 1])  # type: ignore

    axes[2, 1].set_title("Betti Curve")
    axes[2, 1].plot(betti_curves)

    axes[3, 1].set_title("Persistence Landscapes")
    for landscape in persistence_landscapes:
        axes[3, 1].plot(landscape)

    fig.tight_layout()
    plt.gca()
    return axes, axe_0, diagram, fig, grid, gudhi, landscape, plt


if __name__ == "__main__":
    app.run()



from enum import Enum
from typing import Any

import gudhi
import numpy as np
from scipy.stats import gaussian_kde as kde
from sklearn.datasets import make_blobs, make_circles, make_moons
from sklearn.decomposition import PCA
from sklearn.neighbors import KDTree

from topologyx.clustering.unionfind import UnionFind
from topologyx.clustering.utils import plot_clusters, plot_density, plot_persistence


class ClusterStructure(Enum):
    ANISOTROPY = "anisotropy"
    BLOBS = "blobs"
    CIRCLES = "circles"
    MOONS = "moons"
    RANDOM = "random"
    VARIANCES = "variances"


class ClusterGenerator:
    def __init__(
        self,
        structure: ClusterStructure = ClusterStructure.BLOBS,
        n_samples: int = 1500,
        randomize: int = 42,
    ) -> None:
        self.structure = structure
        self.n_samples = n_samples
        self.randomize = randomize

    def generate(self) -> tuple[np.ndarray | None, ...]:
        match self.structure:
            case ClusterStructure.ANISOTROPY:
                x, y = make_blobs(n_samples=self.n_samples, random_state=self.randomize)  # type: ignore
                offset = [[0.60834549, -0.63667341], [-0.40887718, 0.85253229]]

                return (np.dot(x, offset), y)

            case ClusterStructure.VARIANCES:
                return make_blobs(
                    n_samples=self.n_samples,
                    cluster_std=[1.0, 2.5, 0.5],
                    random_state=self.randomize,
                )

            case ClusterStructure.CIRCLES:
                return make_circles(n_samples=self.n_samples, factor=0.5, noise=0.05)

            case ClusterStructure.MOONS:
                return make_moons(n_samples=self.n_samples, noise=0.05)

            case ClusterStructure.RANDOM:
                generator = np.random.default_rng(seed=self.randomize)

                return generator.random((self.n_samples, 2)), None

            case _:
                return make_blobs(n_samples=self.n_samples, random_state=self.randomize)


class TomatoClustering:
    def __init__(self, x: np.ndarray, y: np.ndarray | None = None) -> None:
        self.x = x
        self.y = y

        self.x_plan = PCA(n_components=2).fit_transform(x) if x.shape[1] > 2 else x

        self.estimate_clusters()

    def estimate_density(self, n_bins: int = 100, visualize: bool = False) -> Any:
        kernel_density = kde(self.x.T)
        vector_density = kernel_density(np.vstack(([*self.x.T])))

        if visualize:
            plot_density(self.x_plan, n_bins, vector_density)

        return vector_density

    def estimate_clusters(self, n_neighbors: int = 6, visualize: bool = False) -> None:
        vector_density = self.estimate_density(visualize=visualize)

        self.simplex = gudhi.SimplexTree()  # type: ignore
        self.kd_tree = KDTree(self.x, metric="euclidean")

        for index in range(self.x.shape[0]):
            self.simplex.insert([index], filtration=-vector_density[index])

            neighbors = self.kd_tree.query(
                [self.x[index]], n_neighbors, return_distance=False
            )[0][1:]

            for neighbor_index in neighbors:
                self.simplex.insert(
                    [index, neighbor_index],
                    filtration=np.mean(
                        [-vector_density[index], -vector_density[neighbor_index]]
                    ),
                )

        self.simplex.persistence()

        if visualize:
            plot_persistence(self.simplex.persistence())

    def define_clusters(
        self,
        vertexes: np.ndarray,
        n_simplexes: dict[int, np.ndarray],
        n_neighbors: int,
        tau: float = 1e-2,
    ) -> UnionFind:
        union_find = UnionFind()

        for vertex in vertexes:
            vertex_index = np.nonzero(vertexes == vertex)[0][0]

            if neighbors := [
                neighbor
                for neighbor in self.kd_tree.query(
                    [self.x[vertex]], n_neighbors, return_distance=False
                )[0][1:]
                if np.nonzero(vertexes == neighbor)[0][0] < vertex_index
            ]:
                parent = neighbors[
                    np.asarray(
                        [n_simplexes[neighbor] for neighbor in neighbors]
                    ).argmax()
                ]

                union_find.union(parent, vertex)

                for neighbor in neighbors:
                    root = union_find.find(neighbor)

                    if (
                        root != parent
                        and min(n_simplexes[parent], n_simplexes[root])  # type: ignore
                        < n_simplexes[vertex] + tau
                    ):
                        union_find.union(parent, root)
                        parent = union_find.find(root)

            else:
                union_find.insert_objects([vertex])

        return union_find

    def fit_predict(
        self,
        n_clusters: int | None = None,
        tau: float = 1e-2,
        n_neighbors: int = 6,
        visualize: bool = False,
    ) -> np.ndarray:
        if not hasattr(self, "simplex"):
            self.estimate_clusters(n_neighbors=n_neighbors, visualize=visualize)

        _vertexes, filtrations = [], []

        for simplex, filtration in self.simplex.get_filtration():
            if len(simplex) == 1:
                _vertexes.append(simplex[0])
                filtrations.append(-filtration)

        vertexes = np.asarray(_vertexes)
        n_simplexes = dict(zip(vertexes, np.asarray(filtrations)))

        neighbors_counter = n_neighbors
        union_find = self.define_clusters(vertexes, n_simplexes, n_neighbors)

        if n_clusters is not None:
            while len(np.unique(list(union_find.pointers.values()))) > n_clusters:
                neighbors_counter += 2
                union_find = self.define_clusters(
                    vertexes, n_simplexes, neighbors_counter, tau=tau
                )

        self.clusters, self.centroids = [], []

        indexes = np.asarray(list(union_find.indexes_to_objects.values()))
        pointers = np.asarray(list(union_find.pointers.values()))

        for pointer in np.unique(pointers):
            self.clusters.append(indexes[np.nonzero(pointers == pointer)[0]])
            self.centroids.append(union_find.indexes_to_objects[int(pointer)])

        if visualize:
            plot_clusters(self.x, self.clusters, self.centroids)

        return pointers



import numpy as np

from topologyx.clustering import TomatoClustering


def test_tomato_initialization(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    assert tomato.x is not None
    assert tomato.x.shape == (8, 2)
    assert tomato.y is None
    assert hasattr(tomato, "simplex")
    assert hasattr(tomato, "kd_tree")


def test_tomato_initialization_with_labels(
    blobs_data: tuple[np.ndarray, np.ndarray],
) -> None:
    x, y = blobs_data

    tomato = TomatoClustering(x, y)

    assert tomato.y is not None
    assert len(tomato.y) == len(tomato.x)


def test_estimate_density(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    density = tomato.estimate_density()

    assert density is not None
    assert len(density) == len(simple_2d_data)
    assert all(d > 0 for d in density)


def test_estimate_clusters_creates_simplex_tree(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    tomato.estimate_clusters(n_neighbors=3)

    assert hasattr(tomato, "simplex")
    assert hasattr(tomato, "kd_tree")


def test_fit_predict_returns_cluster_assignments(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    pointers = tomato.fit_predict(n_clusters=2)

    assert pointers is not None
    assert len(pointers) == len(simple_2d_data)


def test_fit_predict_with_n_clusters(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    pointers = tomato.fit_predict(n_clusters=2)
    unique_clusters = np.unique(pointers)

    assert len(unique_clusters) <= 2


def test_fit_predict_creates_clusters_and_centroids(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    tomato.fit_predict(n_clusters=2)

    assert hasattr(tomato, "clusters")
    assert hasattr(tomato, "centroids")
    assert len(tomato.clusters) == len(tomato.centroids)


def test_tomato_blobs_clustering(blobs_data: tuple[np.ndarray, np.ndarray]) -> None:
    x, y = blobs_data

    tomato = TomatoClustering(x, y)
    pointers = tomato.fit_predict(n_clusters=3)

    unique_clusters = np.unique(pointers)
    assert len(unique_clusters) <= 3
    assert len(tomato.clusters) <= 3


def test_tomato_moons_clustering(moons_data: tuple[np.ndarray, np.ndarray]) -> None:
    x, y = moons_data

    tomato = TomatoClustering(x, y)
    pointers = tomato.fit_predict(n_clusters=2)

    unique_clusters = np.unique(pointers)
    assert len(unique_clusters) <= 2


def test_tomato_circles_clustering(circles_data: tuple[np.ndarray, np.ndarray]) -> None:
    x, y = circles_data

    tomato = TomatoClustering(x, y)
    pointers = tomato.fit_predict(n_clusters=2, n_neighbors=15, tau=0.5)

    unique_clusters = np.unique(pointers)
    assert len(unique_clusters) <= 2


def test_fit_predict_with_custom_tau(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    pointers = tomato.fit_predict(n_clusters=2, tau=0.1)

    assert pointers is not None
    assert len(pointers) == len(simple_2d_data)


def test_fit_predict_with_custom_n_neighbors(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    pointers = tomato.fit_predict(n_clusters=2, n_neighbors=3)

    assert pointers is not None
    assert len(pointers) == len(simple_2d_data)


def test_define_clusters_returns_unionfind(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    _vertexes, filtrations = [], []
    for simplex, filtration in tomato.simplex.get_filtration():
        if len(simplex) == 1:
            _vertexes.append(simplex[0])
            filtrations.append(-filtration)

    vertexes = np.asarray(_vertexes)
    n_simplexes = dict(zip(vertexes, np.asarray(filtrations)))

    union_find = tomato.define_clusters(vertexes, n_simplexes, n_neighbors=3)

    assert union_find is not None
    assert len(union_find.objects_to_indexes) > 0


def test_x_plan_is_2d_for_high_dimensional_data() -> None:
    high_dim_data = np.random.default_rng(42).random((100, 10))

    tomato = TomatoClustering(high_dim_data)

    assert tomato.x_plan.shape == (100, 2)


def test_x_plan_unchanged_for_2d_data(simple_2d_data: np.ndarray) -> None:
    tomato = TomatoClustering(simple_2d_data)

    np.testing.assert_array_equal(tomato.x_plan, simple_2d_data)



from topologyx.clustering.unionfind import UnionFind


def test_empty_initialization() -> None:
    uf = UnionFind()

    assert uf.weights == {}
    assert uf.pointers == {}
    assert uf.indexes_to_objects == {}
    assert uf.objects_to_indexes == {}


def test_find_creates_new_element() -> None:
    uf = UnionFind()

    result = uf.find("a")

    assert result == "a"
    assert "a" in uf.objects_to_indexes
    assert uf.weights[0] == 1
    assert uf.pointers[0] == 0


def test_find_returns_same_element_for_singleton() -> None:
    uf = UnionFind()
    uf.find("a")

    result = uf.find("a")

    assert result == "a"


def test_find_with_integer_objects() -> None:
    uf = UnionFind()

    result = uf.find(42)

    assert result == 42
    assert 42 in uf.objects_to_indexes


def test_find_with_tuple_objects() -> None:
    uf = UnionFind()

    result = uf.find((1, 2))

    assert result == (1, 2)
    assert (1, 2) in uf.objects_to_indexes


def test_union_two_elements() -> None:
    uf = UnionFind()
    uf.find("a")
    uf.find("b")

    uf.union("a", "b")

    assert uf.find("a") == uf.find("b")


def test_union_preserves_larger_set_as_root() -> None:
    uf = UnionFind()
    uf.find("a")
    uf.find("b")
    uf.find("c")
    uf.union("a", "b")

    uf.union("a", "c")

    root = uf.find("a")
    assert uf.find("b") == root
    assert uf.find("c") == root


def test_union_idempotent() -> None:
    uf = UnionFind()
    uf.find("a")
    uf.find("b")
    uf.union("a", "b")

    uf.union("a", "b")

    assert uf.find("a") == uf.find("b")


def test_union_three_sets() -> None:
    uf = UnionFind()
    for i in range(6):
        uf.find(i)

    uf.union(0, 1)
    uf.union(2, 3)
    uf.union(4, 5)

    assert uf.find(0) == uf.find(1)
    assert uf.find(2) == uf.find(3)
    assert uf.find(4) == uf.find(5)
    assert uf.find(0) != uf.find(2)
    assert uf.find(0) != uf.find(4)
    assert uf.find(2) != uf.find(4)


def test_union_merges_sets() -> None:
    uf = UnionFind()
    for i in range(4):
        uf.find(i)

    uf.union(0, 1)
    uf.union(2, 3)
    uf.union(1, 2)

    root = uf.find(0)
    assert uf.find(1) == root
    assert uf.find(2) == root
    assert uf.find(3) == root


def test_insert_multiple_objects() -> None:
    uf = UnionFind()

    uf.insert_objects(["a", "b", "c"])

    assert "a" in uf.objects_to_indexes
    assert "b" in uf.objects_to_indexes
    assert "c" in uf.objects_to_indexes
    assert len(uf.objects_to_indexes) == 3


def test_insert_objects_creates_singletons() -> None:
    uf = UnionFind()

    uf.insert_objects([1, 2, 3])

    assert uf.find(1) == 1
    assert uf.find(2) == 2
    assert uf.find(3) == 3


def test_path_compression_on_find() -> None:
    uf = UnionFind()
    for i in range(5):
        uf.find(i)

    uf.union(0, 1)
    uf.union(1, 2)
    uf.union(2, 3)
    uf.union(3, 4)

    root = uf.find(4)

    idx_4 = uf.objects_to_indexes[4]
    assert uf.pointers[idx_4] == uf.objects_to_indexes[root]


def test_weighted_union_attaches_smaller_to_larger() -> None:
    uf = UnionFind()
    for i in range(10):
        uf.find(i)

    for i in range(1, 5):
        uf.union(0, i)

    uf.find(9)
    uf.union(0, 9)

    root = uf.find(0)
    assert uf.find(9) == root


def test_union_same_element() -> None:
    uf = UnionFind()
    uf.find("a")

    uf.union("a", "a")

    assert uf.find("a") == "a"


def test_many_elements() -> None:
    uf = UnionFind()
    n = 1000

    for i in range(n):
        uf.find(i)

    for i in range(0, n - 1, 2):
        uf.union(i, i + 1)

    for i in range(0, n - 1, 2):
        assert uf.find(i) == uf.find(i + 1)


def test_chain_unions() -> None:
    uf = UnionFind()
    n = 100

    for i in range(n):
        uf.find(i)

    for i in range(n - 1):
        uf.union(i, i + 1)

    root = uf.find(0)
    for i in range(n):
        assert uf.find(i) == root
